#include <unistd.h>  // declares read()
#include <stdlib.h>  // declares exit()
#include <stdio.h>   // declares fprintf()
#include <string.h>  // declares strerror()
#include <errno.h>   // declares errno


/**************************
 * Error-handling function
 **************************/
void unix_error(char *msg) /* unix-style error */
{
   fprintf(stderr, "%s: %s (errno = %d)\n", msg, strerror(errno), errno);
   exit(0);
}


/**********************************
 * Robust I/O routines
 **********************************/
#define RIO_BUFSIZE 8192
typedef struct {
   int rio_fd;                /* descriptor for this internal buf */
   int rio_cnt;               /* unread bytes in internal buf */
   char *rio_bufptr;          /* next unread byte in internal buf */
   char rio_buf[RIO_BUFSIZE]; /* internal buffer */
} rio_t;


void rio_readinitb(rio_t *rp, int fd)
{
   rp->rio_fd = fd;
   rp->rio_cnt = 0;
   rp->rio_bufptr = rp->rio_buf;
}


static ssize_t rio_read(rio_t *rp, char *usrbuf, size_t n)
{
   int cnt;

   while (rp->rio_cnt <= 0)
   {  /* refill if buf is empty */
      rp->rio_cnt = read(rp->rio_fd, rp->rio_buf, sizeof(rp->rio_buf));
      if (rp->rio_cnt < 0)
      {
         if (errno != EINTR) /* interrupted by sig handler return */
            return -1;
      }
      else if (rp->rio_cnt == 0)  /* EOF */
         return 0;
      else
         rp->rio_bufptr = rp->rio_buf; /* reset buffer ptr */
   }

   /* Copy min(n, rp->rio_cnt) bytes from internal buf to user buf */
   cnt = n;
   if (rp->rio_cnt < n)
   cnt = rp->rio_cnt;
   memcpy(usrbuf, rp->rio_bufptr, cnt);
   rp->rio_bufptr += cnt;
   rp->rio_cnt -= cnt;
   return cnt;
}


ssize_t rio_readnb(rio_t *rp, void *usrbuf, size_t n)
{
   size_t nleft = n;
   ssize_t nread;
   char *bufp = usrbuf;

   while (nleft > 0)
   {
      if ((nread = rio_read(rp, bufp, nleft)) < 0)
      {
         if (errno == EINTR) /* interrupted by sig handler return */
            nread = 0;       /* call read() again */
         else
            return -1;       /* errno set by read() */
      }
      else
         if (nread == 0)
            break;           /* EOF */
      nleft -= nread;
      bufp += nread;
   }
   return (n - nleft);       /* return >= 0 */
}


/**********************************
 * Wrappers for robust I/O routines
 **********************************/
ssize_t Rio_readnb(rio_t *rp, void *usrbuf, size_t n)
{
   ssize_t rc;

   if ((rc = rio_readnb(rp, usrbuf, n)) < 0)
      unix_error("Rio_readnb error");
   return rc;
}

void Rio_readinitb(rio_t *rp, int fd)
{
   rio_readinitb(rp, fd);
}


/**********************************
 * main()
 **********************************/
int main()
{
   rio_t rio;
   char c;
   int counter = 0;

   Rio_readinitb(&rio, STDIN_FILENO);

   while(Rio_readnb(&rio, &c, 1) != 0)
      counter++;

   printf("%d characters read from standard input", counter);

   exit(0);
}
